/*********   ECE 4760 FINAL PROJECT ****************
*                     GPS    LOGGER               *
*               Jeff Melville jsm66               *
*             Matt Kuzdeba mpk28                  *
***************************************************/

#include <inttypes.h>
#include <avr/io.h>
#include <avr/interrupt.h>
#include <stdio.h>
#include <util/delay.h>
#include <string.h>
#include <stddef.h>
#include "Lab6.h"
#include "nmea.h"
#include "fatfs/ff.h"
#include "lcd_lib.h"
#include <avr/pgmspace.h>

//set up ASSERT
#ifndef __ASSERT_USE_STDERR
#define __ASSERT_USE_STDERR
#include <assert.h>
#endif

//#define DEBUG

const int8_t LCD_initialize[] PROGMEM = "LCD Initialized\0";


//set up UART
//also defines putchar and getchar
#include "uart.h"
FILE uart_str = FDEV_SETUP_STREAM(uart_putchar, uart_getchar, _FDEV_SETUP_RW);

#define nmeaInterval 500  //Interval for NMEA process task, in ms
#define lcdInterval 500  //Interval to update LCD, ms
#define diskInterval 10  //Interval for disk_timerproc (NEEDS to be 10ms)
#define scheduleInterval 250 //Scheduler for interval logging, needs to be 250ms

#define LCD_BUF_SIZE 17
char lcd_buffer[LCD_BUF_SIZE]; //buffer for creating output for LCD

volatile unsigned int nmeaTime; //NMEA process task timer
volatile unsigned int lcdTime;  //LCD update task timer
volatile char scheduleTime;   //Interval logging timer
volatile char diskTime;      //disk_timerproc timer
volatile unsigned int trigTime; //trigger debounce timer

#define TRIGGER_SPACE 1001  //minimum time between triggers, in ms

#define DEFAULT_LOG_INTERVAL 30  //default interval for interval logging (seconds)
volatile unsigned int quarterSecs; //4*number of seconds since last log
volatile char writeFlag; //Flag to initiate a log write
char triggerEnable;  //Setting to enable trigger based logging
char intervalEnable; //Setting to enable interval based logging
unsigned int logInterval; //timed log interval, in seconds


FIL logFile;
FIL* logPtr;
FATFS fs;
FATFS* fsPtr;

void disk_timerproc(); //declaration for disk_timerproc(), in ff.c
//time0 overflow ISR... 1ms timebase
ISR (TIMER0_COMPA_vect) {
	// increment system time in msec
	//time++;
	//Decrement timers
	if (nmeaTime>0) --nmeaTime;
	if (lcdTime>0) --lcdTime;
	if (diskTime>0) --diskTime;
	if (scheduleTime>0) --scheduleTime;
	if (diskTime == 0) {
			diskTime = diskInterval;
			//needs to run async, call from interrupt
			disk_timerproc();
	}
	if (trigTime > 0) --trigTime;
	IncrementSysTime();
}

// trigger on rising edge of analog comparator
ISR (ANALOG_COMP_vect) {
	
	//Only consider this a valid trigger if
    //triggering is enabled, and the timeout since the last one
    //has expired
	if (triggerEnable && trigTime == 0) {
		//reset the timer
		trigTime = TRIGGER_SPACE;
		#ifdef DEBUG
		fprintf(stdout, "TRIGGER!\r\n");
		#endif
		//Signal to write the log
		writeFlag = 1;
		//Reset the interval timer in case both modes are enabled
		quarterSecs = 0;
	}
}


// UART character ready ISR
ISR (USART0_RX_vect) {
	char r_char = UDR0;
	//pass the character to the NMEA parser
	NmeaPutChar(r_char);
}

//Update the current system in days/hours/minutes at a 1ms 
void IncrementSysTime() {
	if (++msec == 60000) {
		msec = 0;
		minutes++;
	}
	if (minutes == 60) {
		minutes = 0;
		hours++;
	}
	if (hours == 24) {
		hours = 0;
		days++;
	}
}

DWORD get_fattime ()
{
	/*
	bit31:25
	Year from 1980 (0..127)
	
	bit24:21
	Month (1..12)
	
	bit20:16
	Day in month(1..31)
	
	bit15:11
	Hour (0..23)
	
	bit10:5
	Minute (0..59)
	
	bit4:0
	Second / 2 (0..29) 
	*/
	
	//For FAT RTC, use GPS time if available
    // Otherwise, make an estimate based on uptime,
    // which will be wildly incorrect.
	if (GPSIsValid()) {
		return	  ((DWORD)(gpsTime.year - 1980) << 25)
			| ((DWORD)gpsTime.month << 21)
			| ((DWORD)gpsTime.day << 16)
			| ((DWORD)gpsTime.hour << 11)
			| ((DWORD)gpsTime.minute << 5)
			| ((DWORD)(gpsTime.second >> 1));

	}

	else {

	return	  ((DWORD)(2009 - 1980) << 25)
			| ((DWORD)4 << 21)
			| ((DWORD)(days+1) << 16)
			| ((DWORD)hours << 11)
			| ((DWORD)minutes << 5)
			| ((DWORD)(msec >> 10) >> 1); 
	// div 1024 is close enough to div 1000 since the time is
    // pretty made up in this case anyway

	}
}

//Main loop stuff
int main() {
	initialize();
	#ifdef DEBUG
	fprintf(stdout, "LOGGER IS UP\r\n");
	#endif
	init_fs();
	load_settings();
	#ifdef DEBUG
	fprintf(stdout, "triggerEnable: %d  intervalEnable: %d  logInterval: %d\r\n", triggerEnable, intervalEnable, logInterval);
	#endif
	open_log();
	init_lcd();
	LCDclr();
	
	while (1) {
		
		if (nmeaTime == 0) {
			nmeaTime = nmeaInterval; 
			NmeaProcessTask();  //in nmea.c
		}
	
		if (lcdTime == 0) {
			lcdTime = lcdInterval;
			LcdUpdateTask();
		}
		
		if (intervalEnable && scheduleTime == 0) {
			scheduleTime = scheduleInterval;
			if (++quarterSecs == (logInterval << 2)) {
				quarterSecs = 0;
				writeFlag = 1;
			}
		}

		if (writeFlag) {
			writeFlag = 0;
			WriteLog();		
		}
	}

}


void LcdUpdateTask() {
	//Update the LCD, using position information if available
    //Otherwise, use "NO GPS FIX"
	if (GPSIsValid()) {
		LCDGotoXY(1,0);
		snprintf(lcd_buffer, LCD_BUF_SIZE, "Lat: %4.3f", GPSGetLatitude());
		LCDstring(lcd_buffer, strlen(lcd_buffer));
		LCDGotoXY(1,1);
		snprintf(lcd_buffer, LCD_BUF_SIZE, "Lon: %4.3f", GPSGetLongitude());
		LCDstring(lcd_buffer, strlen(lcd_buffer));
	}
	else {
		LCDclr();
		snprintf(lcd_buffer, LCD_BUF_SIZE, "NO GPS FIX");
		LCDGotoXY(2,0);
		LCDstring(lcd_buffer, strlen(lcd_buffer));
	}

}


void initialize() {
	
	//Do any GPS init related things in nmea.c
	initGPS();
	DDRD |= (1 << PORTD2);
	
	//init the uart, in uart.c
	uart_init();
	stdout = stdin = stderr = &uart_str;
	//set up timer 0
	OCR0A = 249;  //clear after 250 counts
	TIMSK0 = (1<<OCIE0A); //turn on timer 0 compare match ISR

	//turn on timer0 clear on match
	TCCR0A = (1 << WGM01);
	//timer 0 prescalar to 64
	TCCR0B = 3;
	
	//configure analog comparator interrupt
    //for rising edge, using external input for the 
    //negative comparator input (B2, B3)
	ADCSRB &= ~(1 << ACME);
	ACSR = (1 << ACIE) | (1 << ACIS1) | (1 << ACIS0);
 
	sei(); 
  
	//enable RX char interrupt to asyncrhonously receive
	UCSR0B |= (1<<RXCIE0);
}

void init_fs() {
	FRESULT result;
	
	//initialize the pointers
	fsPtr = &fs;
	logPtr = &logFile;
	//set up the file system
	result = f_mount(0, fsPtr);
	if (result) {
		#ifdef DEBUG
		fprintf(stdout, "Filesystem mount FAILED!\r\n");
		#endif
		return;
	}
}

void load_settings() {
	//This function reads gps_set.txt and configures the following parameters:
   // triggerEnable, intervalEnable, logInterval
   // Defaults are 1,1,30
	
	FRESULT result;
	FIL settingsFile;
	FIL* settingsPtr = &settingsFile;
	char lineBuf[32];
	
	result = f_open(settingsPtr, "gps_set.txt", FA_READ);
	if (result != 0) {
		//Couldn't open file, Load defaults and bail
		triggerEnable = 1;
		intervalEnable = 1;
		logInterval = DEFAULT_LOG_INTERVAL;
		#ifdef DEBUG
		fprintf(stdout, "Using default settings %d result\r\n", result);
		#endif
		return;
	}

	//Settings file existed, read each line and look for config options
	while(f_gets(lineBuf, 32, settingsPtr)) {
		if (strncasecmp(lineBuf, "interval", strlen("interval")) == 0) {
			intervalEnable = 1;
			sscanf(lineBuf, "%*s %d", &logInterval);
		}
		else if (strncasecmp(lineBuf, "trigger", strlen("trigger")) == 0) {
			triggerEnable = 1;
		}
	}
	f_close(settingsPtr);
}
	
				
				

void open_log() {
	FRESULT result;
	char filePath[16];
	unsigned int counter = 0;
	//Try to create a new log file. It will fail if the log exists, keep incrementing log file counter until we succeed
	do {
		snprintf(filePath, 16, "gps%03d.txt", counter);
		result = f_open(logPtr, filePath, FA_WRITE | FA_CREATE_NEW);
	} while (result == FR_EXIST && ++counter < 1000);
	if (result) {
		#ifdef DEBUG
		fprintf(stdout, "GPS log creation FAILED: %d\r\n", result);
		#endif
	}
	else {
		#ifdef DEBUG
		fprintf(stdout, "GPS Log Created: %s", filePath);
		#endif
	}
}

void WriteLog() {
	FRESULT result;
	char* str;
	unsigned int length;
	unsigned int bytesWrote;
	
	//turn on the LED
    //indicates a write, and shows that filesystem is in
	// a critical section
	PORTD &= ~(1 << PORTD2);
	
	//write GGA, GSA, and RMC, accumulating errors if any

	str = GetNmea_GGA();
	length = strlen(str);
	result = f_write(logPtr, str, strlen(str), &bytesWrote);
	
	str = GetNmea_GSA();
	length = strlen(str);
	result |= f_write(logPtr, str, strlen(str), &bytesWrote);
	
	str = GetNmea_RMC();
	length = strlen(str);
	result |= f_write(logPtr, str, strlen(str), &bytesWrote);
	
	#ifdef DEBUG
	if (result | (bytesWrote != length)) {
		fprintf(stdout, "Log write error. Result %d, Tried %d, Wrote %d\r\n", result, length, bytesWrote);
	}
	else {
		fprintf(stdout, "Log write succeeded.\r\n");
	}
	#endif
	//sync the filesystem (flush buffers) so it will power down cleanly
	f_sync(logPtr);
	PORTD |= (1 << PORTD2); //LED off
}

//set up the LCD
void init_lcd(void) {
	LCDinit();	//initialize the display
	LCDcursorOFF();
	LCDclr();				//clear the display

	LCDGotoXY(0,0);
	CopyStringtoLCD(LCD_initialize, 0, 0);
}	
